@@ -4,30 +4,12 @@
 //
 // This module parses the very complicated exif structures.
 //
-// Matthias Wandel,  Dec 1999 - Dec 2002 
+// Matthias Wandel
 //--------------------------------------------------------------------------
-#include <stdio.h>
-#include <stdlib.h>
-#include <math.h>
-#include <string.h>
-#include <time.h>
-#include <errno.h>
-#include <ctype.h>
-
-
-#ifdef _WIN32
-    #include <sys/utime.h>
-#else
-    #include <utime.h>
-    #include <sys/types.h>
-    #include <unistd.h>
-    #include <errno.h>
-    #include <limits.h>
-#endif
-
 #include "jhead.h"
 
-static unsigned char * LastExifRefd;
+#include <math.h>
+
 static unsigned char * DirWithThumbnailPtrs;
 static double FocalplaneXRes;
 static double FocalplaneUnits;
@@ -35,8 +17,9 @@ static int ExifImageWidth;
 static int MotorolaOrder = 0;
 
 // for fixing the rotation.
-static void * OrientationPtr;
-static int    OrientationNumFormat; 
+static void * OrientationPtr[2];
+static int    OrientationNumFormat[2];
+int NumOrientations = 0;
 
 typedef struct {
     unsigned short Tag;
@@ -129,6 +112,7 @@ const int BytesPerFormat[] = {0,1,1,2,4,8,1,1,2,4,8,4,8};
 #define TAG_WHITEBALANCE       0xa403
 #define TAG_DIGITALZOOMRATIO   0xA404
 #define TAG_FOCALLENGTH_35MM   0xa405
+#define TAG_DISTANCE_RANGE     0xa40c
 
 static const TagTable_t TagTable[] = {
   { 0x001,   "InteropIndex"},
@@ -234,8 +218,9 @@ static const TagTable_t TagTable[] = {
 
 #define TAG_TABLE_SIZE  (sizeof(TagTable) / sizeof(TagTable_t))
 
+
 //--------------------------------------------------------------------------
-// Convert a 16 bit unsigned value from file's native byte order
+// Convert a 16 bit unsigned value to file's native byte order
 //--------------------------------------------------------------------------
 static void Put16u(void * Short, unsigned short PutValue)
 {
@@ -274,6 +259,24 @@ int Get32s(void * Long)
     }
 }
 
+//--------------------------------------------------------------------------
+// Convert a 32 bit unsigned value to file's native byte order
+//--------------------------------------------------------------------------
+void Put32u(void * Value, unsigned PutValue)
+{
+    if (MotorolaOrder){
+        ((uchar *)Value)[0] = (uchar)(PutValue>>24);
+        ((uchar *)Value)[1] = (uchar)(PutValue>>16);
+        ((uchar *)Value)[2] = (uchar)(PutValue>>8);
+        ((uchar *)Value)[3] = (uchar)PutValue;
+    }else{
+        ((uchar *)Value)[0] = (uchar)PutValue;
+        ((uchar *)Value)[1] = (uchar)(PutValue>>8);
+        ((uchar *)Value)[2] = (uchar)(PutValue>>16);
+        ((uchar *)Value)[3] = (uchar)(PutValue>>24);
+    }
+}
+
 //--------------------------------------------------------------------------
 // Convert a 32 bit unsigned value from file's native byte order
 //--------------------------------------------------------------------------
@@ -287,23 +290,35 @@ unsigned Get32u(void * Long)
 //--------------------------------------------------------------------------
 void PrintFormatNumber(void * ValuePtr, int Format, int ByteCount)
 {
-    switch(Format){
-        case FMT_SBYTE:
-        case FMT_BYTE:      printf("%02x",*(uchar *)ValuePtr);            break;
-        case FMT_USHORT:    printf("%d",Get16u(ValuePtr));                break;
-        case FMT_ULONG:     
-        case FMT_SLONG:     printf("%d",Get32s(ValuePtr));                break;
-        case FMT_SSHORT:    printf("%hd",(signed short)Get16u(ValuePtr)); break;
-        case FMT_URATIONAL:
-        case FMT_SRATIONAL: 
-           printf("%d/%d",Get32s(ValuePtr), Get32s(4+(char *)ValuePtr)); break;
+    int s,n;
+
+    for(n=0;n<16;n++){
+        switch(Format){
+            case FMT_SBYTE:
+            case FMT_BYTE:      printf("%02x",*(uchar *)ValuePtr); s=1;  break;
+            case FMT_USHORT:    printf("%d",Get16u(ValuePtr)); s=2;      break;
+            case FMT_ULONG:     
+            case FMT_SLONG:     printf("%d",Get32s(ValuePtr)); s=4;      break;
+            case FMT_SSHORT:    printf("%hd",(signed short)Get16u(ValuePtr)); s=2; break;
+            case FMT_URATIONAL:
+            case FMT_SRATIONAL: 
+               printf("%d/%d",Get32s(ValuePtr), Get32s(4+(char *)ValuePtr)); 
+               s = 8;
+               break;
+
+            case FMT_SINGLE:    printf("%f",(double)*(float *)ValuePtr); s=8; break;
+            case FMT_DOUBLE:    printf("%f",*(double *)ValuePtr);        s=8; break;
+            default: 
+                printf("Unknown format %d:", Format);
+                return;
+        }
+        ByteCount -= s;
+        if (ByteCount <= 0) break;
+        printf(", ");
+        ValuePtr = (void *)((char *)ValuePtr + s);
 
-        case FMT_SINGLE:    printf("%f",(double)*(float *)ValuePtr);   break;
-        case FMT_DOUBLE:    printf("%f",*(double *)ValuePtr);          break;
-        default: 
-            printf("Unknown format %d:", Format);
-            
     }
+    if (n >= 16) printf("...");
 }
 
 
@@ -342,6 +357,9 @@ double ConvertAnyFormat(void * ValuePtr, int Format)
         // Not sure if this is correct (never seen float used in Exif format)
         case FMT_SINGLE:    Value = (double)*(float *)ValuePtr;      break;
         case FMT_DOUBLE:    Value = *(double *)ValuePtr;             break;
+
+        default:
+            ErrNonfatal("Illegal format code %d",Format,0);
     }
     return Value;
 }
@@ -383,7 +401,11 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 return;
             }
         }
-        if (DirEnd > LastExifRefd) LastExifRefd = DirEnd;
+        if (DumpExifMap){
+            printf("Map: %05d-%05d: Directory\n",DirStart-OffsetBase, DirEnd+4-OffsetBase);
+        }
+
+
     }
 
     if (ShowTags){
@@ -407,6 +429,11 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             continue;
         }
 
+        if ((unsigned)Components > 0x10000){
+            ErrNonfatal("Illegal number of components %d for tag %04x", Components, Tag);
+            continue;
+        }
+
         ByteCount = Components * BytesPerFormat[Format];
 
         if (ByteCount > 4){
@@ -419,18 +446,19 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 continue;
             }
             ValuePtr = OffsetBase+OffsetVal;
+
+            if (OffsetVal > ImageInfo.LargestExifOffset){
+                ImageInfo.LargestExifOffset = OffsetVal;
+            }
+
+            if (DumpExifMap){
+                printf("Map: %05d-%05d:   Data for tag %04x\n",OffsetVal, OffsetVal+ByteCount, Tag);
+            }
         }else{
             // 4 bytes or less and value is in the dir entry itself
             ValuePtr = DirEntry+8;
         }
 
-        if (LastExifRefd < ValuePtr+ByteCount){
-            // Keep track of last byte in the exif header that was actually referenced.
-            // That way, we know where the discardable thumbnail data begins.
-            LastExifRefd = ValuePtr+ByteCount;
-
-        }
-
         if (Tag == TAG_MAKER_NOTE){
             if (ShowTags){
                 printf("%s    Maker note: ",IndentString);
@@ -456,6 +484,14 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
 
             // Show tag value.
             switch(Format){
+                case FMT_BYTE:
+                    if(ByteCount>1){
+                        printf("%.*ls\n", ByteCount/2, (wchar_t *)ValuePtr);
+                    }else{
+                        PrintFormatNumber(ValuePtr, Format, ByteCount);
+                        printf("\n");
+                    }
+                    break;
 
                 case FMT_UNDEFINED:
                     // Undefined is typically an ascii string.
@@ -517,7 +553,8 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                     ErrNonfatal("More than %d date fields!  This is nuts", MAX_DATE_COPIES, 0);
                     break;
                 }
-                ImageInfo.DateTimePointers[ImageInfo.numDateTimeTags++] = (char *)ValuePtr;
+                ImageInfo.DateTimeOffsets[ImageInfo.numDateTimeTags++] = 
+                    (char *)ValuePtr - (char *)OffsetBase;
                 break;
 
 
@@ -598,18 +635,22 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 break;
 
             case TAG_ORIENTATION:
-                if (OrientationPtr){
-                    // Already have orientation.  The second orientation is likely
-                    // to be that of the thumbnail image.
+                if (NumOrientations >= 2){
+                    // Can have another orientation tag for the thumbnail, but if there's
+                    // a third one, things are stringae.
+                    ErrNonfatal("More than two orientation tags!",0,0);
                     break;
                 }
-                ImageInfo.Orientation = (int)ConvertAnyFormat(ValuePtr, Format);
-                OrientationPtr = ValuePtr;
-                OrientationNumFormat = Format;
-                if (ImageInfo.Orientation < 1 || ImageInfo.Orientation > 8){
+                OrientationPtr[NumOrientations] = ValuePtr;
+                OrientationNumFormat[NumOrientations] = Format;
+                if (NumOrientations == 0){
+                    ImageInfo.Orientation = (int)ConvertAnyFormat(ValuePtr, Format);
+                }
+                if (ImageInfo.Orientation < 0 || ImageInfo.Orientation > 8){
                     ErrNonfatal("Undefined rotation value %d", ImageInfo.Orientation, 0);
                     ImageInfo.Orientation = 0;
                 }
+                NumOrientations += 1;
                 break;
 
             case TAG_EXIF_IMAGELENGTH:
@@ -692,6 +733,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
 
             case TAG_THUMBNAIL_LENGTH:
                 ThumbnailSize = (unsigned)ConvertAnyFormat(ValuePtr, Format);
+                ImageInfo.ThumbnailSizeOffset = ValuePtr-OffsetBase;
                 break;
 
             case TAG_EXIF_OFFSET:
@@ -731,6 +773,12 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 // computing it from sensor geometry and actual focal length.
                 ImageInfo.FocalLength35mmEquiv = (unsigned)ConvertAnyFormat(ValuePtr, Format);
                 break;
+
+            case TAG_DISTANCE_RANGE:
+                // Three possible standard values:
+                //   1 = macro, 2 = close, 3 = distant
+                ImageInfo.DistanceRange = (int)ConvertAnyFormat(ValuePtr, Format);
+                break;
         }
     }
 
@@ -738,7 +786,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
     {
         // In addition to linking to subdirectories via exif tags, 
         // there's also a potential link to another directory at the end of each
-        // directory.  this has got to be the result of a comitee!
+        // directory.  this has got to be the result of a committee!
         unsigned char * SubdirStart;
         unsigned Offset;
 
@@ -761,13 +809,22 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                         ProcessExifDir(SubdirStart, OffsetBase, ExifLength, NestingLevel+1);
                     }
                 }
+                if (Offset > ImageInfo.LargestExifOffset){
+                    ImageInfo.LargestExifOffset = Offset;
+                }
             }
         }else{
             // The exif header ends before the last next directory pointer.
         }
     }
 
-    if (ThumbnailSize && ThumbnailOffset){
+    if (ThumbnailOffset){
+        ImageInfo.ThumbnailAtEnd = FALSE;
+
+        if (DumpExifMap){
+            printf("Map: %05d-%05d: Thumbnail\n",ThumbnailOffset, ThumbnailOffset+ThumbnailSize);
+        }
+
         if (ThumbnailOffset <= ExifLength){
             if (ThumbnailSize > ExifLength-ThumbnailOffset){
                 // If thumbnail extends past exif header, only save the part that
@@ -778,7 +835,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
 
             }
             // The thumbnail pointer appears to be valid.  Store it.
-            ImageInfo.ThumbnailPointer = OffsetBase + ThumbnailOffset;
+            ImageInfo.ThumbnailOffset = ThumbnailOffset;
             ImageInfo.ThumbnailSize = ThumbnailSize;
 
             if (ShowTags){
@@ -788,6 +845,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
     }
 }
 
+
 //--------------------------------------------------------------------------
 // Process a EXIF marker
 // Describes all the drivel that most digital cameras include...
@@ -799,8 +857,7 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
     FocalplaneXRes = 0;
     FocalplaneUnits = 0;
     ExifImageWidth = 0;
-    OrientationPtr = NULL;
-
+    NumOrientations = 0;
 
     if (ShowTags){
         printf("Exif header %d bytes long\n",length);
@@ -835,16 +892,28 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
 
     FirstOffset = Get32u(ExifSection+12);
     if (FirstOffset < 8 || FirstOffset > 16){
-        // I used to ensure this was set to 8 (website I used indicated its 8)
-        // but PENTAX Optio 230 has it set differently, and uses it as offset. (Sept 11 2002)
+        // Usually set to 8, but other values valid too.
         ErrNonfatal("Suspicious offset of first IFD value",0,0);
     }
 
-    LastExifRefd = ExifSection;
     DirWithThumbnailPtrs = NULL;
 
+
     // First directory starts 16 bytes in.  All offset are relative to 8 bytes in.
-    ProcessExifDir(ExifSection+8+FirstOffset, ExifSection+8, length-6, 0);
+    ProcessExifDir(ExifSection+8+FirstOffset, ExifSection+8, length-8, 0);
+
+    ImageInfo.ThumbnailAtEnd = ImageInfo.ThumbnailOffset >= ImageInfo.LargestExifOffset ? TRUE : FALSE;
+
+    if (DumpExifMap){
+        unsigned a,b;
+        printf("Map: %05d- End of exif\n",length-8);
+        for (a=0;a<length-8;a+= 10){
+            printf("Map: %05d ",a);
+            for (b=0;b<10;b++) printf(" %02x",*(ExifSection+8+a+b));
+            printf("\n");
+        }
+    }
+
 
     // Compute the CCD width, in millimeters.
     if (FocalplaneXRes != 0){
@@ -860,65 +929,178 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
             ImageInfo.FocalLength35mmEquiv = (int)(ImageInfo.FocalLength/ImageInfo.CCDWidth*36 + 0.5);
         }
     }
+}
 
-    if (ShowTags){
-        printf("Non settings part of Exif header: %d bytes\n",ExifSection+length-LastExifRefd);
+
+//--------------------------------------------------------------------------
+// Create minimal exif header - just date and thumbnail pointers,
+// so that date and thumbnail may be filled later.
+//--------------------------------------------------------------------------
+void create_EXIF(void)
+{
+    char Buffer[256];
+
+    unsigned short NumEntries;
+    int DataWriteIndex;
+    int DirIndex;
+
+    MotorolaOrder = 0;
+
+    memcpy(Buffer+2, "Exif\0\0II",8);
+    Put16u(Buffer+10, 0x2a);
+
+    DataWriteIndex = 16;
+    Put32u(Buffer+12, DataWriteIndex-8); // first IFD offset.  Means start 16 bytes in.
+
+    {
+        DirIndex = DataWriteIndex;
+        NumEntries = 2;
+        DataWriteIndex += 2 + NumEntries*12 + 4;
+
+        Put16u(Buffer+DirIndex, NumEntries); // Number of entries
+        DirIndex += 2;
+        // Entries go here....
+        {
+            // Date/time entry
+            Put16u(Buffer+DirIndex, TAG_DATETIME);         // Tag
+            Put16u(Buffer+DirIndex + 2, FMT_STRING);       // Format
+            Put32u(Buffer+DirIndex + 4, 20);               // Components
+            Put32u(Buffer+DirIndex + 8, DataWriteIndex-8); // Pointer or value.
+            DirIndex += 12;
+
+            if (ImageInfo.numDateTimeTags){
+                // If we had a pre-existing exif header, use time from that.
+                memcpy(Buffer+DataWriteIndex, ImageInfo.DateTime, 19);
+                Buffer[DataWriteIndex+19] = '\0';
+            }else{
+                // Oterwise, use the file's timestamp.
+                FileTimeAsString(Buffer+DataWriteIndex);
+            }
+            DataWriteIndex += 20;
+        }
+        {
+            // Link to exif dir entry
+            Put16u(Buffer+DirIndex, TAG_EXIF_OFFSET);      // Tag
+            Put16u(Buffer+DirIndex + 2, FMT_ULONG);        // Format
+            Put32u(Buffer+DirIndex + 4, 1);                // Components
+            Put32u(Buffer+DirIndex + 8, DataWriteIndex-8); // Pointer or value.
+            DirIndex += 12;
+        }
+
+        // End of directory - contains optional link to continued directory.
+        Put32u(Buffer+DirIndex, 0);
+    }
+
+
+    {
+        DirIndex = DataWriteIndex;
+        NumEntries = 2;
+        DataWriteIndex += 2 + NumEntries*12 + 4;
+
+        Put16u(Buffer+DirIndex, NumEntries); // Number of entries
+        DirIndex += 2;
+        {
+            // Link to exif dir entry
+            Put16u(Buffer+DirIndex, TAG_THUMBNAIL_OFFSET);         // Tag
+            Put16u(Buffer+DirIndex + 2, FMT_ULONG);       // Format
+            Put32u(Buffer+DirIndex + 4, 1);               // Components
+            Put32u(Buffer+DirIndex + 8, DataWriteIndex-8); // Pointer or value.
+            DirIndex += 12;
+        }
+
+        {
+            // Link to exif dir entry
+            Put16u(Buffer+DirIndex, TAG_THUMBNAIL_LENGTH);         // Tag
+            Put16u(Buffer+DirIndex + 2, FMT_ULONG);       // Format
+            Put32u(Buffer+DirIndex + 4, 1);               // Components
+            Put32u(Buffer+DirIndex + 8, 0); // Pointer or value.
+            DirIndex += 12;
+        }
+
+        // End of directory - contains optional link to continued directory.
+        Put32u(Buffer+DirIndex, 0);
+    }
+
+    Buffer[0] = (unsigned char)(DataWriteIndex >> 8);
+    Buffer[1] = (unsigned char)DataWriteIndex;
+
+    // Remove old exif section, if there was one.
+    RemoveSectionType(M_EXIF);
+
+    {
+        // Sections need malloced buffers, so do that now, especially because
+        // we now know how big it needs to be allocated.
+        unsigned char * NewBuf = malloc(DataWriteIndex);
+        if (NewBuf == NULL){
+            ErrFatal("Could not allocate memory");
+        }
+        memcpy(NewBuf, Buffer, DataWriteIndex);
+
+        CreateSection(M_EXIF, NewBuf, DataWriteIndex);
+
+        // Re-parse new exif section, now that its in place
+        // otherwise, we risk touching data that has already been freed.
+        process_EXIF(NewBuf, DataWriteIndex);
     }
 }
 
 //--------------------------------------------------------------------------
 // Cler the rotation tag in the exif header to 1.
-// Note: The thumbnail is NOT rotated.  That would be really hard, especially
-// stuffing it back into the exif header, because its size might change.
 //--------------------------------------------------------------------------
 const char * ClearOrientation(void)
 {
-    if (OrientationPtr == NULL) return NULL;
-    
-    switch(OrientationNumFormat){
-        case FMT_SBYTE:
-        case FMT_BYTE:      
-            *(uchar *)OrientationPtr = 1;
-            break;
+    int a;
+    if (NumOrientations == 0) return NULL;
 
-        case FMT_USHORT:    
-            Put16u(OrientationPtr, 1);                
-            break;
+    for (a=0;a<NumOrientations;a++){
+        switch(OrientationNumFormat[a]){
+            case FMT_SBYTE:
+            case FMT_BYTE:      
+                *(uchar *)(OrientationPtr[a]) = 1;
+                break;
 
-        case FMT_ULONG:     
-        case FMT_SLONG:     
-            memset(OrientationPtr, 0, 4);
-            // Can't be bothered to write  generic Put32 if I only use it once.
-            if (MotorolaOrder){
-                ((uchar *)OrientationPtr)[3] = 1;
-            }else{
-                ((uchar *)OrientationPtr)[0] = 1;
-            }
-            break;
+            case FMT_USHORT:    
+                Put16u(OrientationPtr[a], 1);                
+                break;
 
-        default:
-            return NULL;
+            case FMT_ULONG:     
+            case FMT_SLONG:     
+                memset(OrientationPtr, 0, 4);
+                // Can't be bothered to write  generic Put32 if I only use it once.
+                if (MotorolaOrder){
+                    ((uchar *)OrientationPtr[a])[3] = 1;
+                }else{
+                    ((uchar *)OrientationPtr[a])[0] = 1;
+                }
+                break;
+
+            default:
+                return NULL;
+        }
     }
 
     return OrientTab[ImageInfo.Orientation];
 }
 
 
+
 //--------------------------------------------------------------------------
 // Remove thumbnail out of the exif image.
 //--------------------------------------------------------------------------
-int RemoveThumbnail(unsigned char * ExifSection, unsigned int Length)
+int RemoveThumbnail(unsigned char * ExifSection)
 {
-
-    // Ensure pointers are up to date.
-    {
-        int ShowTagsTemp = ShowTags;
-        ShowTags = FALSE;
-        process_EXIF(ExifSection, Length);
-        ShowTags = ShowTagsTemp;
+    if (!DirWithThumbnailPtrs || 
+        ImageInfo.ThumbnailOffset == 0 || 
+        ImageInfo.ThumbnailSize == 0){
+        // No thumbnail, or already deleted it.
+        return 0;
+    }
+    if (ImageInfo.ThumbnailAtEnd == FALSE){
+        ErrNonfatal("Thumbnail is not at end of header, can't chop it off", 0, 0);
+        return 0;
     }
 
-    if (DirWithThumbnailPtrs){
+    {
         int de;
         int NumDirEntries;
         NumDirEntries = Get16u(DirWithThumbnailPtrs);
@@ -928,24 +1110,21 @@ int RemoveThumbnail(unsigned char * ExifSection, unsigned int Length)
             unsigned char * DirEntry;
             DirEntry = DIR_ENTRY_ADDR(DirWithThumbnailPtrs, de);
             Tag = Get16u(DirEntry);
-            if (Tag == TAG_THUMBNAIL_OFFSET || Tag == TAG_THUMBNAIL_LENGTH){
-                // We remove data out of the exif directory by doing a memmove on the rest
-                // of the directory to close the gap.
-                // It would of course be far better to have a general purpose read/write
-                // implementation of the filesystem in the exif header, but that would
-                // be quite complicated and therefore very error prone.
-                memmove(DirEntry, 
-                        DIR_ENTRY_ADDR(DirWithThumbnailPtrs, de+1),
-                        (NumDirEntries-de-1)*12+4);
-                NumDirEntries -= 1;
-                de -= 1;
+            if (Tag == TAG_THUMBNAIL_LENGTH){
+                // Set length to zero.
+                if (Get16u(DirEntry+2) != FMT_ULONG){
+                    // non standard format encoding.  Can't do it.
+                    ErrNonfatal("Can't remove thumbnail", 0, 0);
+                    return 0;
+                }
+                Put32u(DirEntry+8, 0);
             }                    
         }
-        Put16u(DirWithThumbnailPtrs, (unsigned short)NumDirEntries);
     }
 
     // This is how far the non thumbnail data went.
-    return LastExifRefd - ExifSection;
+    return ImageInfo.ThumbnailOffset+8;
+
 }
 
 
@@ -976,6 +1155,7 @@ int Exif2tm(struct tm * timeptr, char * ExifTime)
     return FALSE; // Wasn't in Exif date format.
 }
 
+
 //--------------------------------------------------------------------------
 // Show the collected image info, displaying camera F-stop and shutter speed
 // in a consistent and legible fashion.
@@ -988,9 +1168,7 @@ void ShowImageInfo(int ShowFileInfo)
 
         {
             char Temp[20];
-            struct tm ts;
-            ts = *localtime(&ImageInfo.FileDateTime);
-            strftime(Temp, 20, "%Y:%m:%d %H:%M:%S", &ts);
+            FileTimeAsString(Temp);
             printf("File date    : %s\n",Temp);
         }
     }
@@ -1021,7 +1199,7 @@ void ShowImageInfo(int ShowFileInfo)
 	            case 0x7: printf(" (Strobe light detected) "); break;
 	            case 0x9: printf(" (manual)"); break;
 	            case 0xd: printf(" (manual, return light not detected)"); break;
-	            case 0xf: printf(" (maual, return light  detected)"); break;
+	            case 0xf: printf(" (manual, return light  detected)"); break;
 	            case 0x19:printf(" (auto)"); break;
 	            case 0x1d:printf(" (auto, return light not detected)"); break;
 	            case 0x1f:printf(" (auto, return light detected)"); break;
@@ -1030,7 +1208,7 @@ void ShowImageInfo(int ShowFileInfo)
 	            case 0x47:printf(" (red eye reduction mode return light  detected)"); break;
 	            case 0x49:printf(" (manual, red eye reduction mode)"); break;
 	            case 0x4d:printf(" (manual, red eye reduction mode, return light not detected)"); break;
-	            case 0x4f:printf(" (red eye reduction mode, red eye reduction mode, return light detected)"); break;
+	            case 0x4f:printf(" (red eye reduction mode, return light detected)"); break;
 	            case 0x59:printf(" (auto, red eye reduction mode)"); break;
 	            case 0x5d:printf(" (auto, red eye reduction mode, return light not detected)"); break;
 	            case 0x5f:printf(" (auto, red eye reduction mode, return light detected)"); break;
@@ -1157,7 +1335,7 @@ void ShowImageInfo(int ShowFileInfo)
             printf("Exposure     : shutter priority (semi-auto)\n");
             break;
         case 5:
-            printf("Exposure     : Creative Program (based towards depth of field\n"); 
+            printf("Exposure     : Creative Program (based towards depth of field)\n"); 
             break;
         case 6:
             printf("Exposure     : Action program (based towards fast shutter speed)\n");
@@ -1176,9 +1354,28 @@ void ShowImageInfo(int ShowFileInfo)
         case 0: // Automatic (not worth cluttering up output for)
             break;
         case 1: printf("Exposure Mode: Manual\n");
+            break;
         case 2: printf("Exposure Mode: Auto bracketing\n");
+            break;
     }
 
+    if (ImageInfo.DistanceRange) {
+        printf("Focus range  : ");
+        switch(ImageInfo.DistanceRange) {
+            case 1:
+                printf("macro");
+                break;
+            case 2:
+                printf("close");
+                break;
+            case 3:
+                printf("distant");
+                break;
+        }
+        printf("\n");
+    }
+
+
 
     if (ImageInfo.Process != M_SOF0){
         // don't show it if its the plain old boring 'baseline' process, but do
@@ -1223,7 +1420,6 @@ void ShowImageInfo(int ShowFileInfo)
         }
         printf("\n");
     }
-    printf("\n");
 }
 
 
@@ -1237,7 +1433,11 @@ void ShowConciseImageInfo(void)
     printf(" %dx%d",ImageInfo.Width, ImageInfo.Height);
 
     if (ImageInfo.ExposureTime){
-        printf(" (1/%d)",(int)(0.5 + 1/ImageInfo.ExposureTime));
+        if (ImageInfo.ExposureTime <= 0.5){
+            printf(" (1/%d)",(int)(0.5 + 1/ImageInfo.ExposureTime));
+        }else{
+            printf(" (%1.1f)",ImageInfo.ExposureTime);
+        }
     }
 
     if (ImageInfo.ApertureFNumber){
