@@ -3,17 +3,6 @@
 //
 // Matthias Wandel,  Dec 1999 - Dec 2002 
 //--------------------------------------------------------------------------
-#include <stdio.h>
-#include <stdlib.h>
-#include <math.h>
-#include <string.h>
-#include <time.h>
-#include <errno.h>
-#include <ctype.h>
-
-#ifndef _WIN32
-    #include <limits.h>
-#endif
 #include "jhead.h"
 
 #define MAX_GPS_TAG 0x1e
@@ -64,14 +53,12 @@ static const char * GpsTags[MAX_GPS_TAG+1]= {
 //--------------------------------------------------------------------------
 // Process GPS info directory
 //--------------------------------------------------------------------------
-void ProcessGpsInfo(unsigned char * DirStart, int ByteCount, unsigned char * OffsetBase, unsigned ExifLength)
+void ProcessGpsInfo(unsigned char * DirStart, int ByteCountUnused, unsigned char * OffsetBase, unsigned ExifLength)
 {
     int de;
     unsigned a;
     int NumDirEntries;
 
-    ByteCount; // We assume byte count is big enough.
-
     NumDirEntries = Get16u(DirStart);
     #define DIR_ENTRY_ADDR(Start, Entry) (Start+2+12*(Entry))
 
@@ -92,6 +79,11 @@ void ProcessGpsInfo(unsigned char * DirStart, int ByteCount, unsigned char * Off
         unsigned char * DirEntry;
         DirEntry = DIR_ENTRY_ADDR(DirStart, de);
 
+        if (DirEntry+12 > OffsetBase+ExifLength){
+            ErrNonfatal("GPS info directory goes past end of exif",0,0);
+            return;
+        }
+
         Tag = Get16u(DirEntry);
         Format = Get16u(DirEntry+2);
         Components = Get32u(DirEntry+4);
