@@ -2,46 +2,29 @@
 // Program to pull the information out of various types of EXIF digital 
 // camera files and show it in a reasonably consistent way
 //
-// Version 2.6
-//
+// Version 2.8
 //
 // Compiling under Windows:  
-//   Make sure you have microsoft's compiler on the path, then run make.bat
+//   Make sure you have Microsoft's compiler on the path, then run make.bat
 //
-// Dec 1999 - April 2006
+// Dec 1999 - Nov 2007
 //
 // by Matthias Wandel   www.sentex.net/~mwandel
 //--------------------------------------------------------------------------
-#include <stdio.h>
-#include <stdlib.h>
-#include <memory.h>
-#include <stdlib.h>
-#include <string.h>
-#include <time.h>
-#include <sys/stat.h>
-#include <errno.h>
-#include <ctype.h>
+#include "jhead.h"
 
-#define JHEAD_VERSION "2.6"
+#include <sys/stat.h>
+
+#define JHEAD_VERSION "2.8"
 
 // This #define turns on features that are too very specific to 
 // how I organize my photos.  Best to ignore everything inside #ifdef MATTHIAS
 //#define MATTHIAS
 
 #ifdef _WIN32
-    #include <process.h>
     #include <io.h>
-    #include <sys/utime.h>
-#else
-    #include <utime.h>
-    #include <sys/types.h>
-    #include <unistd.h>
-    #include <errno.h>
-    #include <limits.h>
 #endif
 
-#include "jhead.h"
-
 static int FilesMatched;
 static int FileSequence;
 
@@ -59,17 +42,23 @@ static int Exif2FileTime  = FALSE;
 static int DoModify     = FALSE;
 static int DoReadAction = FALSE;
        int ShowTags     = FALSE;    // Do not show raw by default.
+static int Quiet        = FALSE;    // Be quiet on success (like unix programs)
        int DumpExifMap  = FALSE;
 static int ShowConcise  = FALSE;
+static int CreateExifSection = FALSE;
 static char * ApplyCommand = NULL;  // Apply this command to all images.
 static char * FilterModel = NULL;
 static int    ExifOnly    = FALSE;
-static int    PortraitOnly = 0;
+static int    PortraitOnly = FALSE;
 static time_t ExifTimeAdjust = 0;   // Timezone adjust
 static time_t ExifTimeSet = 0;      // Set exif time to a value.
+static char DateSet[11];
+static unsigned DateSetChars = 0;
+static unsigned FileTimeToExif = FALSE;
 
 static int DeleteComments = FALSE;
 static int DeleteExif = FALSE;
+static int DeleteIptc = FALSE;
 static int DeleteUnknown = FALSE;
 static char * ThumbSaveName = NULL; // If not NULL, use this string to make up
                                     // the filename to store the thumbnail to.
@@ -77,7 +66,7 @@ static char * ThumbSaveName = NULL; // If not NULL, use this string to make up
 static char * ThumbInsertName = NULL; // If not NULL, use this string to make up
                                     // the filename to retrieve the thumbnail from.
 
-static int RegenThumbnail = 0;
+static int RegenThumbnail = FALSE;
 
 static char * ExifXferScrFile = NULL;// Extract Exif header from this file, and
                                     // put it into the Jpegs processed.
@@ -89,10 +78,10 @@ static char * CommentSavefileName = NULL; // Save comment to this file.
 static char * CommentInsertfileName = NULL; // Insert comment from this file.
 static char * CommentInsertLiteral = NULL;  // Insert this comment (from command line)
 
-static int AutoRotate = 0;
-static int ZeroRotateTagOnly = 0;
+static int AutoRotate = FALSE;
+static int ZeroRotateTagOnly = FALSE;
 
-static int ShowFileInfo = 1;        // Indicates to show standard file info
+static int ShowFileInfo = TRUE;     // Indicates to show standard file info
                                     // (file name, file size, file date)
 
 
@@ -468,7 +457,9 @@ void RenameAssociated(const char * FileName, char * NewBaseName)
         strcat(NewName, finddata.name+a); // add extension to new name
 
         if (rename(FilePattern, NewName) == 0){
-            printf("%s --> %s\n",FilePattern, NewName);
+            if (!Quiet){
+                printf("%s --> %s\n",FilePattern, NewName);
+            }
         }
 
         next_file:
@@ -699,23 +690,26 @@ static int RegenerateThumbnail(const char * FileName)
     }
 }
 
+//--------------------------------------------------------------------------
+// Set file time as exif time.
+//--------------------------------------------------------------------------
+void FileTimeAsString(char * TimeStr)
+{
+    struct tm ts;
+    ts = *localtime(&ImageInfo.FileDateTime);
+    strftime(TimeStr, 20, "%Y:%m:%d %H:%M:%S", &ts);
+}
+
 //--------------------------------------------------------------------------
 // Do selected operations to one file at a time.
 //--------------------------------------------------------------------------
 void ProcessFile(const char * FileName)
 {
     int Modified = FALSE;
-    ReadMode_t ReadMode = READ_EXIF;
+    ReadMode_t ReadMode = READ_METADATA;
     CurrentFile = FileName;
     FilesMatched = 1; 
 
-    if (DoModify || RenameToDate || Exif2FileTime){
-        if (access(FileName, 2 /*W_OK*/)){
-            printf("Skipping readonly file '%s'\n",FileName);
-            return;
-        }
-    }
-
     ResetJpgfile();
 
     // Start with an empty image information structure.
@@ -735,6 +729,13 @@ void ProcessFile(const char * FileName)
         }
     }
 
+    if (DoModify || RenameToDate || Exif2FileTime){
+        if (access(FileName, 2 /*W_OK*/)){
+            printf("Skipping readonly file '%s'\n",FileName);
+            return;
+        }
+    }
+
     strncpy(ImageInfo.FileName, FileName, PATH_MAX);
 
 
@@ -742,7 +743,7 @@ void ProcessFile(const char * FileName)
         // Applying a command is special - the headers from the file have to be
         // pre-read, then the command executed, and then the image part of the file read.
 
-        if (!ReadJpegFile(FileName, READ_EXIF)) return;
+        if (!ReadJpegFile(FileName, READ_METADATA)) return;
 
         #ifdef MATTHIAS
             if (AutoResize){
@@ -768,7 +769,7 @@ void ProcessFile(const char * FileName)
             }
         }else{
             struct stat dummy;
-            DoCommand(FileName, TRUE);
+            DoCommand(FileName, Quiet ? FALSE : TRUE);
 
             if (stat(FileName, &dummy)){
                 // The file is not there anymore. Perhaps the command
@@ -785,7 +786,7 @@ void ProcessFile(const char * FileName)
         // Make a relative name.
         RelativeName(RelativeExifName, ExifXferScrFile, FileName);
 
-        if(!ReadJpegFile(RelativeExifName, READ_EXIF)) return;
+        if(!ReadJpegFile(RelativeExifName, READ_METADATA)) return;
 
         DiscardAllButExif();    // Don't re-read exif section again on next read.
 
@@ -811,6 +812,17 @@ void ProcessFile(const char * FileName)
     }else{
         if (!(DoModify || DoReadAction) || ShowTags){
             ShowImageInfo(ShowFileInfo);
+
+            {
+                // if IPTC section is present, show it also.
+                Section_t * IptcSection;
+                IptcSection = FindSection(M_IPTC);
+            
+                if (IptcSection){
+                    show_IPTC(IptcSection->Data, IptcSection->Size);
+                }
+            }
+            printf("\n");
         }
     }
 
@@ -820,12 +832,16 @@ void ProcessFile(const char * FileName)
         RelativeName(OutFileName, ThumbSaveName, FileName);
 
         if (SaveThumbnail(OutFileName)){
-            if (OutFileName != "-"){
-                printf("Created: '%s'\n", OutFileName);
-            }
+            printf("Created: '%s'\n", OutFileName);
         }
     }
 
+    if (CreateExifSection){
+        // Make a new minimal exif section
+        create_EXIF();
+        Modified = TRUE;
+    }
+
     if (RegenThumbnail){
         if (RegenerateThumbnail(FileName)){
             Modified = TRUE;
@@ -960,19 +976,29 @@ void ProcessFile(const char * FileName)
         }
     }
 
-
-    if (ExifTimeAdjust || ExifTimeSet){
+    if (ExifTimeAdjust || ExifTimeSet || DateSetChars || FileTimeToExif){
        if (ImageInfo.numDateTimeTags){
             struct tm tm;
             time_t UnixTime;
             char TempBuf[50];
             int a;
             Section_t * ExifSection;
-
             if (ExifTimeSet){
                 // A time to set was specified.
                 UnixTime = ExifTimeSet;
             }else{
+                if (FileTimeToExif){
+                    FileTimeAsString(ImageInfo.DateTime);
+                }
+                if (DateSetChars){
+                    memcpy(ImageInfo.DateTime, DateSet, DateSetChars);
+                    a = 1970;
+                    sscanf(DateSet, "%d", &a);
+                    if (a < 1970){
+                        strcpy(TempBuf, ImageInfo.DateTime);
+                        goto skip_unixtime;
+                    }
+                }
                 // A time offset to adjust by was specified.
                 if (!Exif2tm(&tm, ImageInfo.DateTime)) goto badtime;
 
@@ -984,11 +1010,12 @@ void ProcessFile(const char * FileName)
             tm = *localtime(&UnixTime);
 
             // Print to temp buffer first to avoid putting null termination in destination.
-            // snprintf() would do the trick ,but not available everywhere (like FreeBSD 4.4)
+            // snprintf() would do the trick, hbut not available everywhere (like FreeBSD 4.4)
             sprintf(TempBuf, "%04d:%02d:%02d %02d:%02d:%02d",
                 tm.tm_year+1900, tm.tm_mon+1, tm.tm_mday,
                 tm.tm_hour, tm.tm_min, tm.tm_sec);
-
+            
+skip_unixtime:         
             ExifSection = FindSection(M_EXIF);
 
             for (a = 0; a < ImageInfo.numDateTimeTags; a++) {
@@ -996,6 +1023,7 @@ void ProcessFile(const char * FileName)
                 Pointer = ExifSection->Data+ImageInfo.DateTimeOffsets[a]+8;
                 memcpy(Pointer, TempBuf, 19);
             }
+            memcpy(ImageInfo.DateTime, TempBuf, 19);
 
             Modified = TRUE;
         }else{
@@ -1009,6 +1037,9 @@ void ProcessFile(const char * FileName)
     if (DeleteExif){
         if (RemoveSectionType(M_EXIF)) Modified = TRUE;
     }
+    if (DeleteIptc){
+        if (RemoveSectionType(M_IPTC)) Modified = TRUE;
+    }
     if (DeleteUnknown){
         if (RemoveUnknownSections()) Modified = TRUE;
     }
@@ -1018,7 +1049,7 @@ void ProcessFile(const char * FileName)
         char BackupName[400];
         struct stat buf;
 
-        printf("Modified: %s\n",FileName);
+        if (!Quiet) printf("Modified: %s\n",FileName);
 
         strcpy(BackupName, FileName);
         strcat(BackupName, ".t");
@@ -1069,7 +1100,7 @@ void ProcessFile(const char * FileName)
             if (utime(FileName, &mtime) != 0){
                 printf("Error: Could not change time of file '%s'\n",FileName);
             }else{
-                printf("%s\n",FileName);
+                if (!Quiet) printf("%s\n",FileName);
             }
         }else{
             printf("File '%s' contains no Exif timestamp\n", FileName);
@@ -1095,7 +1126,7 @@ badtime:
 static void Usage (void)
 {
     printf("Jhead is a program for manipulating settings and thumnails in Exif jpeg headers\n"
-           "used by most Digital Cameras.  v"JHEAD_VERSION" Matthias Wandel, April 29 2006.\n"
+           "used by most Digital Cameras.  v"JHEAD_VERSION" Matthias Wandel, Nov 13 2007.\n"
            "http://www.sentex.net/~mwandel/jhead\n"
            "\n");
 
@@ -1109,8 +1140,10 @@ static void Usage (void)
            "             Uses same name mangling as '-st' option\n"
            "  -dc        Delete comment field (as left by progs like Photoshop & Compupic)\n"
            "  -de        Strip Exif section (smaller JPEG file, but lose digicam info)\n"
+           "  -di        Delete IPTC section (from Photoshop, or Picasa)\n"
            "  -du        Delete non image sections except for Exif and comment sections\n"
            "  -purejpg   Strip all unnecessary data from jpeg (combines -dc -de and -du)\n"
+           "  -mkexif    Create new minimal exif section (overwrites pre-existing exif)\n"
            "  -ce        Edit comment field.  Uses environment variable 'editor' to\n"
            "             determine which editor to use.  If editor not set, uses VI\n"
            "             under Unix and notepad with windows\n"
@@ -1120,7 +1153,8 @@ static void Usage (void)
            "  -cl string Insert literal comment string\n"
 
            "\nDATE / TIME MANIPULATION:\n"
-           "  -ft        Set file modification time to Exif time.\n"
+           "  -ft        Set file modification time to Exif time\n"
+           "  -dsft      Set Exif time to file modification time\n"
            "  -n[format-string]\n"
            "             Rename files according to date.  Uses exif date if present, file\n"
            "             date otherwise.  If the optional format-string is not supplied,\n"
@@ -1154,10 +1188,12 @@ static void Usage (void)
            "             To deal with different months and years having different numbers of\n"
            "             days, a simple date-month-year offset would result in unexpected\n"
            "             results.  Instead, the difference is specified as desired date\n"
-           "             minus original date.  Date is specified as yyyy:mmm:dd or as date\n"
-           "             and time in the format yyyy:mmm:dd/hh:mm:ss\n"
+           "             minus original date.  Date is specified as yyyy:mm:dd or as date\n"
+           "             and time in the format yyyy:mm:dd/hh:mm:ss\n"
            "  -ts<time>  Set the Exif internal time to <time>.  <time> is in the format\n"
            "             yyyy:mm:dd-hh:mm:ss\n"
+           "  -ds<date>  Set the Exif internal date.  <date> is in the format YYYY:MM:DD\n"
+           "             or YYYY:MM or YYYY\n"
 
            "\nTHUMBNAIL MANIPULATION:\n"
            "  -dt        Remove exif integral thumbnails.   Typically trims 10k\n"
@@ -1184,6 +1220,7 @@ static void Usage (void)
            "\nOUTPUT VERBOSITY CONTROL:\n"
            "  -h         help (this text)\n"
            "  -v         even more verbose output\n"
+           "  -q         Quiet (no messages on success, like Unix)\n"
            "  -V         Show jhead version\n"
            "  -exifmap   Dump header bytes, annotate.  Pipe thru sort for better viewing\n"
            "  -se        Supress error messages relating to corrupt exif header structure\n"
@@ -1292,12 +1329,16 @@ int main (int argc, char **argv)
         }else if (!strcmp(arg,"-de")){
             DeleteExif = TRUE;
             DoModify = TRUE;
+        }else if (!strcmp(arg,"-di")){
+            DeleteIptc = TRUE;
+            DoModify = TRUE;
         }else if (!strcmp(arg, "-du")){
             DeleteUnknown = TRUE;
             DoModify = TRUE;
         }else if (!strcmp(arg, "-purejpg")){
             DeleteExif = TRUE;
             DeleteComments = TRUE;
+            DeleteIptc = TRUE;
             DeleteUnknown = TRUE;
             DoModify = TRUE;
         }else if (!strcmp(arg,"-ce")){
@@ -1305,19 +1346,23 @@ int main (int argc, char **argv)
             DoModify = TRUE;
         }else if (!strcmp(arg,"-cs")){
             CommentSavefileName = argv[++argn];
-            DoModify = TRUE;
         }else if (!strcmp(arg,"-ci")){
             CommentInsertfileName = argv[++argn];
             DoModify = TRUE;
         }else if (!strcmp(arg,"-cl")){
             CommentInsertLiteral = argv[++argn];
             DoModify = TRUE;
+        }else if (!strcmp(arg,"-mkexif")){
+            CreateExifSection = TRUE;
+            DoModify = TRUE;
 
     // Output verbosity control
         }else if (!strcmp(arg,"-h")){
             Usage();
         }else if (!strcmp(arg,"-v")){
             ShowTags = TRUE;
+        }else if (!strcmp(arg,"-q")){
+            Quiet = TRUE;
         }else if (!strcmp(arg,"-V")){
             printf("Jhead version: "JHEAD_VERSION"   Compiled: "__DATE__"\n");
             exit(0);
@@ -1409,6 +1454,34 @@ int main (int argc, char **argv)
             if (ExifTimeAdjust) ErrFatal("Can only use one of -da or -ta options at once");
             ExifTimeAdjust = NewDate-OldDate;
             DoModify = TRUE;
+        }else if (!memcmp(arg,"-dsft",5)){
+            // Set file time to date/time in exif
+            FileTimeToExif = TRUE;
+            DoModify = TRUE;
+        }else if (!memcmp(arg,"-ds",3)){
+            // Set date feature
+            int a;
+            // Check date validity and copy it.  Could be incompletely specified.
+            strcpy(DateSet, "0000:01:01");
+            for (a=0;arg[a+3];a++){
+                if (isdigit(DateSet[a])){
+                    if (!isdigit(arg[a+3])){
+                        a = 0;
+                        break;
+                    }
+                }else{
+                    if (arg[a+3] != ':'){
+                        a=0;
+                        break;
+                    }
+                }
+                DateSet[a] = arg[a+3];
+            }
+            if (a < 4 || a > 10){
+                ErrFatal("Date must be in format YYYY, YYYY:MM, or YYYY:MM:DD");
+            }
+            DateSetChars = a;
+            DoModify = TRUE;
         }else if (!memcmp(arg,"-ts",3)){
             // Set the exif time.
             // Time must be specified as "yyyy:mm:dd-hh:mm:ss"
