@@ -4,38 +4,16 @@
 //
 // This module handles basic Jpeg file handling
 //
-// Matthias Wandel,  Dec 1999 - Dec 2002 
+// Matthias Wandel
 //--------------------------------------------------------------------------
-#include <stdio.h>
-#include <stdlib.h>
-#include <memory.h>
-#include <stdlib.h>
-#include <string.h>
-#include <time.h>
-#include <sys/stat.h>
-#include <errno.h>
-#include <ctype.h>
-
-#ifdef _WIN32
-    #include <process.h>
-    #include <io.h>
-    #include <sys/utime.h>
-#else
-    #include <utime.h>
-    #include <sys/types.h>
-    #include <unistd.h>
-    #include <errno.h>
-    #include <limits.h>
-#endif
-
 #include "jhead.h"
 
 // Storage for simplified info extracted from file.
 ImageInfo_t ImageInfo;
 
 
-#define MAX_SECTIONS 100
-static Section_t Sections[MAX_SECTIONS];
+static Section_t * Sections = NULL;
+static int SectionsAllocated;
 static int SectionsRead;
 static int HaveAll;
 
@@ -116,6 +94,22 @@ static void process_SOFn (const uchar * Data, int marker)
 }
 
 
+//--------------------------------------------------------------------------
+// Check sections array to see if it needs to be increased in size.
+//--------------------------------------------------------------------------
+void CheckSectionsAllocated(void)
+{
+    if (SectionsRead > SectionsAllocated){
+        ErrFatal("allocation screwup");
+    }
+    if (SectionsRead >= SectionsAllocated){
+        SectionsAllocated += SectionsAllocated/2;
+        Sections = (Section_t *)realloc(Sections, sizeof(Section_t)*SectionsAllocated);
+        if (Sections == NULL){
+            ErrFatal("could not allocate data for entire image");
+        }
+    }
+}
 
 
 //--------------------------------------------------------------------------
@@ -128,7 +122,6 @@ int ReadJpegSections (FILE * infile, ReadMode_t ReadMode)
 
     a = fgetc(infile);
 
-
     if (a != 0xff || fgetc(infile) != M_SOI){
         return FALSE;
     }
@@ -138,16 +131,14 @@ int ReadJpegSections (FILE * infile, ReadMode_t ReadMode)
         int ll,lh, got;
         uchar * Data;
 
-        if (SectionsRead >= MAX_SECTIONS){
-            ErrFatal("Too many sections in jpg file");
-        }
+        CheckSectionsAllocated();
 
         for (a=0;a<7;a++){
             marker = fgetc(infile);
             if (marker != 0xff) break;
 
             if (a >= 6){
-                printf("too many padding bytes\n");
+                fprintf(stderr,"too many padding bytes\n");
                 return FALSE;
             }
         }
@@ -210,6 +201,7 @@ int ReadJpegSections (FILE * infile, ReadMode_t ReadMode)
                         ErrFatal("could not read the rest of the image");
                     }
 
+                    CheckSectionsAllocated();
                     Sections[SectionsRead].Data = Data;
                     Sections[SectionsRead].Size = size;
                     Sections[SectionsRead].Type = PSEUDO_IMAGE_MARKER;
@@ -219,11 +211,11 @@ int ReadJpegSections (FILE * infile, ReadMode_t ReadMode)
                 return TRUE;
 
             case M_EOI:   // in case it's a tables-only JPEG stream
-                printf("No image in jpeg!\n");
+                fprintf(stderr,"No image in jpeg!\n");
                 return FALSE;
 
             case M_COM: // Comment section
-                if (HaveCom || ((ReadMode & READ_EXIF) == 0)){
+                if (HaveCom || ((ReadMode & READ_METADATA) == 0)){
                     // Discard this section.
                     free(Sections[--SectionsRead].Data);
                 }else{
@@ -244,7 +236,7 @@ int ReadJpegSections (FILE * infile, ReadMode_t ReadMode)
                 // Seen files from some 'U-lead' software with Vivitar scanner
                 // that uses marker 31 for non exif stuff.  Thus make sure 
                 // it says 'Exif' in the section before treating it as exif.
-                if ((ReadMode & READ_EXIF) && memcmp(Data+2, "Exif", 4) == 0){
+                if ((ReadMode & READ_METADATA) && memcmp(Data+2, "Exif", 4) == 0){
                     process_EXIF(Data, itemlen);
                 }else{
                     // Discard this section.
@@ -252,6 +244,18 @@ int ReadJpegSections (FILE * infile, ReadMode_t ReadMode)
                 }
                 break;
 
+            case M_IPTC:
+                if (ReadMode & READ_METADATA){
+                    if (ShowTags){
+                        printf("Image cotains IPTC section, %d bytes long\n", itemlen);
+                    }
+                    // Note: We just store the IPTC section.  Its relatively straightforward
+                    // and we don't act on any part of it, so just display it at parse time.
+                }else{
+                    free(Sections[--SectionsRead].Data);
+                }
+                break;
+           
             case M_SOF0: 
             case M_SOF1: 
             case M_SOF2: 
@@ -284,9 +288,11 @@ int ReadJpegSections (FILE * infile, ReadMode_t ReadMode)
 void DiscardData(void)
 {
     int a;
+
     for (a=0;a<SectionsRead;a++){
         free(Sections[a].Data);
     }
+
     memset(&ImageInfo, 0, sizeof(ImageInfo));
     SectionsRead = 0;
     HaveAll = 0;
@@ -310,7 +316,7 @@ int ReadJpegFile(const char * FileName, ReadMode_t ReadMode)
     // Scan the JPEG headers.
     ret = ReadJpegSections(infile, ReadMode);
     if (!ret){
-        printf("Not JPEG: %s\n",FileName);
+        fprintf(stderr,"Not JPEG: %s\n",FileName);
     }
 
     fclose(infile);
@@ -321,29 +327,115 @@ int ReadJpegFile(const char * FileName, ReadMode_t ReadMode)
     return ret;
 }
 
+
 //--------------------------------------------------------------------------
-// Remove exif thumbnail
+// Replace or remove exif thumbnail
 //--------------------------------------------------------------------------
-int TrimExifFunc(void)
+int SaveThumbnail(char * ThumbFileName)
 {
-    int a;
-    for (a=0;a<SectionsRead-1;a++){
-        if (Sections[a].Type == M_EXIF && memcmp(Sections[a].Data+2, "Exif",4)==0){
-            unsigned int NewSize;
-            NewSize = RemoveThumbnail(Sections[a].Data, Sections[a].Size);
-            // Truncate the thumbnail section of the exif.
-            printf("%d bytes removed\n",Sections[a].Size-NewSize);
-            if (Sections[a].Size == NewSize) return FALSE; // Nothing removed.
-            Sections[a].Size = NewSize;
-            Sections[a].Data[0] = (uchar)(NewSize >> 8);
-            Sections[a].Data[1] = (uchar)NewSize;
-            return TRUE;
-        }
+    FILE * ThumbnailFile;
+
+    if (ImageInfo.ThumbnailOffset == 0 || ImageInfo.ThumbnailSize == 0){
+        fprintf(stderr,"Image contains no thumbnail\n");
+        return FALSE;
+    }
+
+    if (strcmp(ThumbFileName, "-") == 0){
+        // A filename of '-' indicates thumbnail goes to stdout.
+        // This doesn't make much sense under Windows, so this feature is unix only.
+        ThumbnailFile = stdout;
+    }else{
+        ThumbnailFile = fopen(ThumbFileName,"wb");
+    }
+
+    if (ThumbnailFile){
+        uchar * ThumbnailPointer;
+        Section_t * ExifSection;
+        ExifSection = FindSection(M_EXIF);
+        ThumbnailPointer = ExifSection->Data+ImageInfo.ThumbnailOffset+8;
+
+        fwrite(ThumbnailPointer, ImageInfo.ThumbnailSize ,1, ThumbnailFile);
+        fclose(ThumbnailFile);
+        return TRUE;
+    }else{
+        ErrFatal("Could not write thumbnail file");
+        return FALSE;
     }
-    // Not an exif image.  Can't remove exif thumbnail.
-    return FALSE;
 }
 
+//--------------------------------------------------------------------------
+// Replace or remove exif thumbnail
+//--------------------------------------------------------------------------
+int ReplaceThumbnail(const char * ThumbFileName)
+{
+    FILE * ThumbnailFile;
+    int ThumbLen, NewExifSize;
+    Section_t * ExifSection;
+    uchar * ThumbnailPointer;
+
+    if (ImageInfo.ThumbnailOffset == 0 || ImageInfo.ThumbnailAtEnd == FALSE){
+        if (ThumbFileName == NULL){
+            // Delete of nonexistent thumbnail (not even pointers present)
+            // No action, no error.
+            return FALSE;
+        }
+
+        // Adding or removing of thumbnail is not possible - that would require rearranging
+        // of the exif header, which is risky, and jhad doesn't know how to do.
+        fprintf(stderr,"Image contains no thumbnail to replace - add is not possible\n");
+        return FALSE;
+    }
+
+    if (ThumbFileName){
+        ThumbnailFile = fopen(ThumbFileName,"rb");
+
+        if (ThumbnailFile == NULL){
+            ErrFatal("Could not read thumbnail file");
+            return FALSE;
+        }
+
+        // get length
+        fseek(ThumbnailFile, 0, SEEK_END);
+
+        ThumbLen = ftell(ThumbnailFile);
+        fseek(ThumbnailFile, 0, SEEK_SET);
+
+        if (ThumbLen + ImageInfo.ThumbnailOffset > 0x10000-20){
+            ErrFatal("Thumbnail is too large to insert into exif header");
+        }
+    }else{
+        if (ImageInfo.ThumbnailSize == 0){
+             return FALSE;
+        }
+
+        ThumbLen = 0;
+        ThumbnailFile = NULL;
+    }
+
+    ExifSection = FindSection(M_EXIF);
+
+    NewExifSize = ImageInfo.ThumbnailOffset+8+ThumbLen;
+    ExifSection->Data = (uchar *)realloc(ExifSection->Data, NewExifSize);
+
+    ThumbnailPointer = ExifSection->Data+ImageInfo.ThumbnailOffset+8;
+
+    if (ThumbnailFile){
+        fread(ThumbnailPointer, ThumbLen, 1, ThumbnailFile);
+        fclose(ThumbnailFile);
+    }
+
+    ImageInfo.ThumbnailSize = ThumbLen;
+
+    Put32u(ExifSection->Data+ImageInfo.ThumbnailSizeOffset+8, ThumbLen);
+
+    ExifSection->Data[0] = (uchar)(NewExifSize >> 8);
+    ExifSection->Data[1] = (uchar)NewExifSize;
+    ExifSection->Size = NewExifSize;
+
+    return TRUE;
+}
+
+
 //--------------------------------------------------------------------------
 // Discard everything but the exif and comment sections.
 //--------------------------------------------------------------------------
@@ -351,27 +443,37 @@ void DiscardAllButExif(void)
 {
     Section_t ExifKeeper;
     Section_t CommentKeeper;
+    Section_t IptcKeeper;
     int a;
 
     memset(&ExifKeeper, 0, sizeof(ExifKeeper));
     memset(&CommentKeeper, 0, sizeof(CommentKeeper));
+    memset(&IptcKeeper, 0, sizeof(IptcKeeper));
 
     for (a=0;a<SectionsRead;a++){
         if (Sections[a].Type == M_EXIF && ExifKeeper.Type == 0){
             ExifKeeper = Sections[a];
         }else if (Sections[a].Type == M_COM && CommentKeeper.Type == 0){
             CommentKeeper = Sections[a];
+        }else if (Sections[a].Type == M_IPTC && IptcKeeper.Type == 0){
+            IptcKeeper = Sections[a];
         }else{
             free(Sections[a].Data);
         }
     }
     SectionsRead = 0;
     if (ExifKeeper.Type){
+        CheckSectionsAllocated();
         Sections[SectionsRead++] = ExifKeeper;
     }
     if (CommentKeeper.Type){
+        CheckSectionsAllocated();
         Sections[SectionsRead++] = CommentKeeper;
     }
+    if (IptcKeeper.Type){
+        CheckSectionsAllocated();
+        Sections[SectionsRead++] = IptcKeeper;
+    }
 }    
 
 //--------------------------------------------------------------------------
@@ -425,7 +527,8 @@ void WriteJpegFile(const char * FileName)
 Section_t * FindSection(int SectionType)
 {
     int a;
-    for (a=0;a<SectionsRead-1;a++){
+
+    for (a=0;a<SectionsRead;a++){
         if (Sections[a].Type == SectionType){
             return &Sections[a];
         }
@@ -484,6 +587,7 @@ int RemoveUnknownSections(void)
             case  M_DQT:
             case  M_DHT:
             case  M_DRI:
+            case  M_IPTC:
                 // keep.
                 a++;
                 break;
@@ -501,29 +605,31 @@ int RemoveUnknownSections(void)
 
 //--------------------------------------------------------------------------
 // Add a section (assume it doesn't already exist) - used for 
-// adding comment sections.
+// adding comment sections and exif sections
 //--------------------------------------------------------------------------
 Section_t * CreateSection(int SectionType, unsigned char * Data, int Size)
 {
     Section_t * NewSection;
     int a;
+    int NewIndex;
+    NewIndex = 2;
+
+    if (SectionType == M_EXIF) NewIndex = 0; // Exif alwas goes first!
 
     // Insert it in third position - seems like a safe place to put 
     // things like comments.
 
-    if (SectionsRead < 2){
+    if (SectionsRead < NewIndex){
         ErrFatal("Too few sections!");
     }
-    if (SectionsRead >= MAX_SECTIONS){
-        ErrFatal("Too many sections!");
-    }
 
-    for (a=SectionsRead;a>2;a--){
+    CheckSectionsAllocated();
+    for (a=SectionsRead;a>NewIndex;a--){
         Sections[a] = Sections[a-1];          
     }
     SectionsRead += 1;
 
-    NewSection = Sections+2;
+    NewSection = Sections+NewIndex;
 
     NewSection->Type = SectionType;
     NewSection->Size = Size;
@@ -538,7 +644,11 @@ Section_t * CreateSection(int SectionType, unsigned char * Data, int Size)
 //--------------------------------------------------------------------------
 void ResetJpgfile(void)
 {
-    memset(&Sections, 0, sizeof(Sections));
+    if (Sections == NULL){
+        Sections = (Section_t *)malloc(sizeof(Section_t)*5);
+        SectionsAllocated = 5;
+    }
+
     SectionsRead = 0;
     HaveAll = 0;
 }
